"""
A helper module for Pet logic.
"""
from dataclasses import dataclass
from typing import Optional

from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.errors import PrintableException, YouFoundABugRewardError
from hopla.hoplalib.hatchery.hatchpotion_data import HatchPotionData
from hopla.hoplalib.zoo.fooddata import FoodData
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petdata import PetData


class InvalidPet(PrintableException):
    """Exception raised when a pet is invalid."""

    def __init__(self, msg: str, *, pet: Optional["Pet"] = None):
        super().__init__(msg)
        self.pet = pet


@dataclass
class Pet:
    """A habitica pet"""

    # pylint: disable=too-many-public-methods
    #################################################################
    # In a way, pylint is right. This class does have a lot of functions.
    # Most of these are oneliners, however we can also use polymorphism.
    # That would make it easier to find functions in this class.
    # We can use a PetFactory that creates QuestPets, MagicPets etc.
    # However, that is really overkill for now and not urgent enough.
    # Therefore, too-many-public-methods is disabled for this class.
    #################################################################

    def __init__(self, pet_name: str, *,
                 feed_status: FeedStatus = FeedStatus(5)):
        if pet_name not in PetData.pet_names:
            raise InvalidPet(f"{pet_name=} is not recognized by hopla.\n"
                             "Potential causes: \n"
                             "* did you spell it correctly?\n"
                             "* Is your pet relatively new? If so, please raise\n"
                             f" an issue at {GlobalConstants.NEW_ISSUE_URL}")

        self.name = pet_name
        self.feed_status = feed_status

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name}: {self.feed_status})"

    @property
    def hatch_potion_name(self) -> Optional[str]:
        """The hatching potion used to hatch the egg this pet came from."""
        if self.name in PetData.rare_pet_names:
            return None

        # This logic might break, but seems to be solid for past few years due to
        # stabling naming convention by the Habitica API developers.
        _, potion = self.name.split("-")
        return potion

    def is_available(self) -> bool:
        """Return True if the feed status says that the pet is available."""
        return self.feed_status.is_pet_available()

    def is_feedable(self) -> bool:
        """Return True if a pet cannot be fed at all."""
        return self.name not in PetData.unfeedable_pet_names

    def has_just_1_favorite_food(self) -> bool:
        """Return True if this pet likes only 1 type of food."""
        return self.name in PetData.only_1favorite_food_pet_names

    def likes_all_food(self) -> bool:
        """Return True if this pet prefers all food."""
        return self.name in PetData.magic_potion_pet_names

    def feed_status_explanation(self) -> str:
        """Explain the feed status of a pet."""
        if self.is_feedable() is False:
            return f"{self.name} is an unfeedable pet."
        if int(self.feed_status) == 0:
            return f"You released {self.name}, so you cannot feed it."

        if int(self.feed_status) == -1:
            return f"You can't feed {self.name=} you only have the mount"

        if int(self.feed_status) == 5:
            # remark: if we really want to know more, we can query the
            #         API to check if we have the self.name in the
            #         Users: jq .data.items.mounts
            return (f"Cannot determine if {self.name=} can be fed. \n"
                    "You either have:\n"
                    "1. Both the pet and mount. In this case, you cannot feed the pet \n"
                    "2. A pet that hasn't been fed but you don't have \n"
                    "   the mount. In this case, you can feed your pet")

        if int(self.feed_status) < 50:
            return f"{self.name=} can be fed"

        raise InvalidPet(f"Did not expect {self.feed_status=}. \n"
                         f"Looks like we failed to validate the Pet properly.",
                         pet=self)  # pragma: no cover

    def favorite_food(self, *,
                      default_value_for_unfeedable: str = "Unfeedable",
                      default_value_for_all_favorite_food: str = "Any") -> str:
        """Return the favorite food of this pet."""
        if self.is_feedable() is False:
            return default_value_for_unfeedable

        if self.name in PetData.magic_potion_pet_names:
            return default_value_for_all_favorite_food

        if self.has_just_1_favorite_food():
            return (FoodData.hatch_potion_favorite_food_mapping
                    .get(self.hatch_potion_name))

        raise InvalidPet(f"Could not find the feed habits of this {self.name=}",
                         pet=self)  # pragma: no cover

    def is_favorite_food(self, food_name: str) -> bool:
        """Return true if 'food_name' is this Pets favorite food."""
        if self.is_feedable() is False:
            return False
        if self.likes_all_food():
            return True
        if self.has_just_1_favorite_food():
            return self.favorite_food() == food_name

        raise InvalidPet(f"Unable to determine if {food_name=} is {self.name=}'s favorite.",
                         pet=self)  # pragma: no cover

    def required_food_items_until_mount(self, food_name: str) -> int:
        """
        Return the number of times food_name needs to be fed
        for this pet to turn into a mount
        """
        is_favorite: bool = self.is_favorite_food(food_name)
        return self.feed_status.required_food_items_to_become_mount(is_favorite)

    def is_generation1_pet(self) -> bool:
        """Return True if this pet is from the generation 1 pet"""
        return self.name in PetData.generation1_pet_names

    def is_quest_pet(self) -> bool:
        """
        Return True if this pet is a quest pet. This doesn't include
        special pets such as world event related pets.
        """
        return self.name in PetData.quest_pet_names

    def is_magic_hatch_pet(self) -> bool:
        """Return True if this pet is hatched from a magic potion."""
        return self.name in PetData.magic_potion_pet_names

    def is_from_drop_hatch_potions(self) -> bool:
        """
        Return True if the pet was hatched from one of the 'ordinary'
        potions. (Such as: Base, Desert, ...).
        """
        return self.hatch_potion_name in HatchPotionData.drop_hatch_potion_names


@dataclass
class Mount:
    """ The model class for mounts."""

    def __init__(self, mount_name: str, *,
                 availability_status: Optional[bool]):
        """
        Create a mount.

        Note that, in contrast to Pet.__init__, currently, there is no name
        correctness checking. Related ticket:
        https://github.com/melvio/hopla/issues/164

        :param mount_name:
        :param availability_status:
        """
        self.name = mount_name
        self._availability_status = availability_status

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name}: {self._availability_status})"

    def is_available(self) -> bool:
        """Return true if the user has the mount right now."""
        return self._availability_status is True


class InvalidPetMountPair(YouFoundABugRewardError):
    """Exception raised when an invalid PetMountPair is created."""


@dataclass
class PetMountPair:
    """A pair of a pet and its mount."""

    def __init__(self, *, pet: Optional[Pet],
                 mount: Optional[Mount]):
        """
        Create a PetMountPair.
        Raise an exception if the names of the pet and mount do not match.
        """
        if pet and mount:
            if pet.name != mount.name:
                msg = f"pet name '{pet.name}' must equal mount name '{mount.name}'"
                raise InvalidPetMountPair(msg)

        self.pet = pet
        self.mount = mount

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pet={self.pet}, mount={self.mount})"

    def can_feed_pet(self) -> bool:
        """Return true if the pet itself is feedable and there is no mount yet."""
        return (
                self.pet_available()
                and self.pet.is_feedable()
                and not self.mount_available()
        )

    def mount_available(self) -> bool:
        """True if the mount is in the pair"""
        return self.mount is not None and self.mount.is_available()

    def pet_available(self) -> bool:
        """True if the pet is in the pair"""
        return self.pet is not None and self.pet.is_available()



#!/usr/bin/env python3
from hopla.hoplalib.buy.buy_controllers import BuyEnchantedArmoireRequest


class TestBuyEnchantedArmoireRequest:
    def test_url(self):
        request = BuyEnchantedArmoireRequest()

        result: str = request.url

        assert result == "https://habitica.com/api/v3/user/buy-armoire"



#!/usr/bin/env python3
import pytest

from hopla.hoplalib.cast.castcontroller import PostCastRequest
from hopla.hoplalib.cast.spellmodel import Spell, SpellData


class TestPostCastRequest:
    @pytest.mark.parametrize("spell_name", SpellData.single_arg_spells)
    def test_url_ok(self, spell_name: str):
        cast_request = PostCastRequest(spell=Spell(spell_name))

        url_result: str = cast_request.url

        assert url_result == f"https://habitica.com/api/v3/user/class/cast/{spell_name}"



#!/usr/bin/env python3
"""TestUtils to get HabiticaUser objects for users."""
from typing import Dict, Optional

from hopla.cli.groupcmds.get_user import HabiticaUser


class UserTestUtil:
    """test utility to create an user"""

    @classmethod
    def user_with_zoo(cls, *,
                      pets: Optional[Dict[str, int]] = None,
                      mounts: Optional[Dict[str, bool]] = None) -> HabiticaUser:
        """Create a simple user to test pet and mount logic"""
        if pets is None:
            pets = {}
        if mounts is None:
            mounts = {}

        return HabiticaUser({"items": {"pets": pets, "mounts": mounts}})

    @classmethod
    def user_with_gp(cls, *,
                     gold: float):
        """Create the simplest user with the specified budget."""
        return HabiticaUser({"stats": {"gp": gold}})



"""
The module with CLI code that handles the `hopla api` group command.
"""
import logging

import click
import requests

from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.http import HabiticaRequest, UrlBuilder
from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.outputformatter import JsonFormatter

log = logging.getLogger()


@click.group()
def api():
    """GROUP for requesting Habitica API metadata."""


class ApiContentRequest:
    """Class that requests a user model from the Habitica API"""

    def __init__(self):
        self.url = UrlBuilder(path_extension="/content").url

    def request_api_content(self) -> requests.Response:
        """Perform the get API content request and return the response"""
        return requests.get(url=self.url, timeout=HabiticaRequest.TIMEOUT)

    def request_api_content_on_fail_exit(self) -> dict:
        """
        Function that requests the habitica API content.
        If the request was successful return the content, else exits.
        """
        api_response: requests.Response = self.request_api_content()
        return get_data_or_exit(api_response)


@api.command()
def content() -> dict:
    """Print detailed information about Habitica's API content.

    \b
    Example
    ----
    # Tip: The content is long, you can try piping it to an editor or pager.
    $ hopla api content | vim -
    $ hopla api content | less -

    \b
    # get content information about
    # the [Ruby Rapport quest](https://habitica.fandom.com/wiki/Ruby_Rapport)
    $ hopla api content | jq ".quests.ruby"


    [API-docs](https://habitica.com/apidoc/#api-Content-ContentGet)
    \f
    :return:
    """
    log.debug("hopla api content")

    content_data: dict = ApiContentRequest().request_api_content_on_fail_exit()
    content_as_json: str = JsonFormatter(content_data).format_with_double_quotes()
    click.echo(content_as_json)
    return content_data


valid_model_names = click.Choice(
    choices=["user", "group", "challenge", "tag", "habit", "daily", "todo", "reward"],
    case_sensitive=False
)


@api.command()
@click.argument("model_name", type=valid_model_names)
def model(model_name: str) -> dict:
    """Print the specified Habitica API datamodel.

    \b
    Example:
    ---
    # Get the 'group' datamodel
    $ hopla api model group

    [apidocs](https://habitica.com/apidoc/#api-Meta-GetUserModelPaths)

    \f
    :param model_name: The particular data model
    :return:
    """
    log.debug(f"hopla api model name={model_name}")

    url_builder = UrlBuilder(path_extension=f"/models/{model_name}/paths")
    response = requests.get(
        url=url_builder.url,
        timeout=HabiticaRequest.TIMEOUT
    )
    model_data = get_data_or_exit(response)

    click.echo(JsonFormatter(model_data).format_with_double_quotes())
    return model_data


@api.command()
def version() -> dict:
    """print the version string of the Habitica API (e.g. v3)

    \f
    :return The habitica API version string. (e.g. v3, v4)
    """
    log.debug("hopla api version")
    api_version = {"version": GlobalConstants.HABITICA_API_VERSION}
    click.echo(JsonFormatter(api_version).format_with_double_quotes())
    return api_version  # .appVersion maybe also interesting?


@api.command()
def status() -> dict:
    """Print the Habitica API availability status.


    \f
    :return: The API status (expected: "status": "up")
    """
    log.debug("hopla api status")

    url = UrlBuilder(path_extension="/status").url
    response = requests.get(url=url, timeout=HabiticaRequest.TIMEOUT)
    status_data = get_data_or_exit(response)

    click.echo(JsonFormatter(status_data).format_with_double_quotes())
    return status_data



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla hatch-all` command.
"""
import logging
import sys
from typing import Any, Callable, Dict, Iterator, List

import click
from requests import Response

from hopla.hoplalib import hopla_option
from hopla.hoplalib.hatchery.eggmodels import EggCollection
from hopla.hoplalib.hatchery.hatchalgorithms import HatchPlan, HatchPlanMaker
from hopla.hoplalib.hatchery.hatchcontroller import HatchRequester
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotionCollection
from hopla.hoplalib.throttling import RateLimitingAwareThrottler
from hopla.hoplalib.user.usercontroller import HabiticaUserRequest
from hopla.hoplalib.user.usermodels import HabiticaUser
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petmodels import Pet

log = logging.getLogger()


@click.command()
@hopla_option.no_interactive_option()
def hatch_all(no_interactive: bool) -> None:
    """Hatch all the available eggs.

    \b
    hopla hatch-all - Print the possible hatchings, and, if the user
                      confirms, hatch all the available eggs
                      with all the available hatching potions.

    \b
    Examples:
    ----
    # Show hatching list, ask for confirmation, optionally hatch eggs.
    $ hopla hatch-all
    > A Treeling egg will be hatched by a Zombie potion.
    > A Robot egg will be hatched by a Desert potion.
    > Do you wish to proceed? [y/N]: yes
    Successfully hatched a Treeling-Zombie.
    Successfully hatched a Robot-Desert.

    \b
    # Hatch all available eggs without asking for confirmation.
    $ hopla hatch-all --yes
    Successfully hatched a Treeling-Zombie.
    Successfully hatched a Robot-Desert.

    """
    log.debug(f"hopla hatch-all {no_interactive=}")
    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    eggs = EggCollection(user.get_eggs())
    potions = HatchPotionCollection(user.get_hatch_potions())
    pets: List[Pet] = to_pet_list(user.get_pets())

    plan_maker = HatchPlanMaker(
        egg_collection=eggs, hatch_potion_collection=potions, pets=pets
    )
    plan: HatchPlan = plan_maker.make_plan()

    if plan.is_empty():
        click.echo(
            "The hatch plan is empty. Do you have enough eggs and hatching potions?\n"
            "Exiting"
        )
        sys.exit(1)

    if no_interactive is True:
        _hatch_eggs_without_confirmation(plan)
    else:
        _ask_for_confirmation_and_maybe_hatch(plan)


def _ask_for_confirmation_and_maybe_hatch(plan: HatchPlan) -> None:
    plan_text: str = plan.format_plan()
    user_confirmed: bool = click.confirm(text=plan_text + "Do you wish to proceed?")
    if user_confirmed is True:
        _hatch_eggs_without_confirmation(plan)
    else:
        click.echo("No eggs were hatched.")


def _hatch_eggs_without_confirmation(plan: HatchPlan) -> None:
    """Hatch all the eggs. Print the result to the terminal.
    Warning: this function does not ask for confirmation.
    """
    # Feel free to refactor this such that we don't iterate over
    # the plan twice.
    api_requests: List[Callable[[], Response]] = [
        HatchRequester(item.egg.name, item.potion.name).post_hatch_egg_request for item in plan
    ]

    throttler = RateLimitingAwareThrottler(api_requests)
    api_responses: Iterator[Response] = throttler.perform_and_yield_response()
    for item in plan:
        response: Response = next(api_responses)
        response_json: Dict[str, Any] = response.json()
        pet_name = f"{item.egg.name}-{item.potion.name}"
        if response_json["success"] is True:
            click.echo(f"Successfully hatched a {pet_name}.")
        else:
            click.echo(f"Failed to hatch {pet_name}:\n"
                       f"{response_json['error']}: {response_json['message']}")


def to_pet_list(pets: Dict[str, int]) -> List[Pet]:
    """Helper method that takes a pet_dict and returns a List[Pet]."""
    return [
        Pet(name, feed_status=FeedStatus(n)) for (name, n) in pets.items()
    ]



#!/usr/bin/env python3
from datetime import datetime, timedelta
from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner, Result

from hopla.cli.add.todo import todo
from hopla.hoplalib.tasks.taskmodel import HabiticaTodo


class MockAddTodoRequest:
    def __init__(self, mocked_response: dict):
        self.mocked_response = mocked_response

    def post_add_todo_request(self) -> dict:
        return self.mocked_response


class TestTodoCliCommand:
    @patch("hopla.cli.add.todo.AddTodoRequest")
    def test_todo_name_only_task(self,
                                 todo_request_mock: MagicMock,
                                 todo_name: str,
                                 mock_todo_response: dict):
        todo_request_mock.return_value = MockAddTodoRequest(mock_todo_response)
        runner = CliRunner()
        result: Result = runner.invoke(todo, [todo_name])

        todo_request_mock.assert_called_with(
            habitica_todo=HabiticaTodo(difficulty="easy",
                                       todo_name=todo_name)
        )
        assert result.exit_code == 0

    @patch("hopla.cli.add.todo.AddTodoRequest")
    def test_todo_no_task_name(self,
                               todo_request_mock: MagicMock,
                               mock_todo_response: dict):
        todo_request_mock.return_value = MockAddTodoRequest(mock_todo_response)
        runner = CliRunner()
        user_input = "cure rhabdomyolysis"
        result: Result = runner.invoke(todo, input=user_input)

        todo_request_mock.assert_called_with(
            habitica_todo=HabiticaTodo(difficulty="easy",
                                       todo_name=user_input)
        )
        assert result.exit_code == 0

    @patch("hopla.cli.add.todo.AddTodoRequest")
    def test_todo_today_option(self,
                               todo_request_mock: MagicMock,
                               todo_name: str,
                               mock_todo_response: dict):
        todo_request_mock.return_value = MockAddTodoRequest(mock_todo_response)
        runner = CliRunner()
        result: Result = runner.invoke(todo, [todo_name, "--today"])

        todo_request_mock.assert_called_with(
            habitica_todo=HabiticaTodo(difficulty="easy",
                                       due_date=datetime.today(),
                                       todo_name=todo_name)
        )
        assert result.exit_code == 0

    @patch("hopla.cli.add.todo.AddTodoRequest")
    def test_todo_tomorrow_option(self,
                                  todo_request_mock: MagicMock,
                                  todo_name: str,
                                  mock_todo_response: dict):
        todo_request_mock.return_value = MockAddTodoRequest(mock_todo_response)
        runner = CliRunner()
        result: Result = runner.invoke(todo, [todo_name, "--tomorrow"])

        tomorrow: datetime = datetime.today() + timedelta(days=1)
        todo_request_mock.assert_called_with(
            habitica_todo=HabiticaTodo(difficulty="easy",
                                       due_date=tomorrow,
                                       todo_name=todo_name)
        )
        assert result.exit_code == 0

    @pytest.mark.parametrize("cmd_due_date,expected_datetime", [
        ("2021-01-02", datetime(year=2021, month=1, day=2)),
        ("1969-12-10", datetime(year=1969, month=12, day=10)),
        ("01-05-2042", datetime(year=2042, month=5, day=1)),
        ("07-11-3012", datetime(year=3012, month=11, day=7))
    ])
    @patch("hopla.cli.add.todo.AddTodoRequest")
    def test_todo_due_date_option(self,
                                  todo_request_mock: MagicMock,
                                  cmd_due_date: str,
                                  expected_datetime: datetime,
                                  todo_name: str,
                                  mock_todo_response: dict):
        todo_request_mock.return_value = MockAddTodoRequest(mock_todo_response)
        runner = CliRunner()
        date_arg = "2021-01-02"
        result: Result = runner.invoke(todo, [todo_name, "--due-date", date_arg])

        todo_request_mock.assert_called_with(
            habitica_todo=HabiticaTodo(difficulty="easy",
                                       due_date=datetime(day=2, month=1, year=2021),
                                       todo_name=todo_name)
        )
        assert result.exit_code == 0

    @pytest.fixture()
    def todo_name(self) -> str:
        return "my_todo"

    @pytest.fixture
    def mock_todo_response(self, todo_name: str) -> dict:
        return {
            "challenge": {},
            "group": {
                "approval": {"required": False, "approved": False, "requested": False},
                "assignedUsers": [], "sharedCompletion": "singleCompletion"
            },
            "completed": False,
            "collapseChecklist": False,
            "type": "todo",
            "notes": "",
            "tags": [],
            "value": 0,
            "priority": 1,
            "attribute": "str",
            "byHabitica": False,
            "text": todo_name,
            "checklist": [],
            "date": "2021-10-03T00:00:00.000Z",
            "reminders": [],
            "_id": "cc00e66e-8dcc-4218-b675-3119ebda00c6",
            "createdAt": "2021-10-02T07:55:02.893Z",
            "updatedAt": "2021-10-02T07:55:02.893Z",
            "userId": "79213ab8-31e9-42b4-b7fa-9d89b0114319",
            "id": "cc78326e-8dcc-4298-b675-3d000bda00c6"
        }



#!/usr/bin/env python3
from typing import Callable

from hopla.hoplalib.hopla_option import no_interactive_option


class TestNoInteractiveOption:
    def test_no_interactive_option(self):
        f = no_interactive_option()
        assert isinstance(f, Callable)



#!/usr/bin/env python3
from unittest.mock import MagicMock, patch
from click.testing import CliRunner, Result

from hopla.cli.get_group import HabiticaGroupRequest, get_group


class MockGroupResponse:
    def __init__(self, json, status_code: int = 200):
        self._json = json
        self.status_code = status_code

    def json(self):
        return self._json


class TestHabiticaGroupRequest:

    def test_path_party(self):
        party_request = HabiticaGroupRequest()
        assert party_request.path == "/groups/party"

    def test_path_uuid(self):
        comrades_uuid = "2ff9822b-27f2-4774-98da-db349b57a38e"
        comrades_request = HabiticaGroupRequest(comrades_uuid)
        assert comrades_request.path == f"/groups/{comrades_uuid}"

    @patch("hopla.cli.get_group.HabiticaGroupRequest.default_headers")
    @patch("hopla.cli.get_group.requests.get")
    def test_get_group_request(self, mock_get: MagicMock,
                               headers: MagicMock):
        party_request = HabiticaGroupRequest()

        mock_get.return_value = MockGroupResponse(
            json={"success": True, "data": get_test_party_dict()}
        )

        # No need for credentials, were mocking the API call
        headers.return_value = {"mock": "headers"}
        party_request.get_group_request()

        mock_get.assert_called_once()
        assert mock_get.call_args.url.endswith("/groups/party")

    @patch("hopla.cli.get_group.HabiticaGroupRequest.default_headers")
    @patch("hopla.cli.get_group.requests.get")
    def test_get_group_or_exit_ok(self,
                                  mock_get: MagicMock,
                                  headers: MagicMock):
        legends_uuid = "52f49529-58c1-4020-a59b-8bb8579e941f"
        legends_request = HabiticaGroupRequest()

        mock_get.return_value = MockGroupResponse(
            json={"success": True, "data": get_test_party_dict()}
        )

        # No need for credentials, were mocking the API call
        headers.return_value = {"mock": "headers"}
        legends_request.get_group_request()

        mock_get.assert_called_once()
        assert mock_get.call_args.url.endswith(f"/groups/{legends_uuid}")


class TestGetGroupCli:

    @patch("hopla.cli.get_group.HabiticaGroupRequest.default_headers")
    @patch("hopla.cli.get_group.requests.get")
    def test_get_group_ok(self,
                          mock_get: MagicMock,
                          headers: MagicMock):
        """A successful hopla get-group call."""
        party_data = get_test_party_dict()
        mock_get.return_value = MockGroupResponse(
            json={"success": True, "data": party_data}
        )

        # No need for credentials, were mocking the API call
        headers.return_value = {"mock": "headers"}

        runner = CliRunner()
        result: Result = runner.invoke(get_group)

        assert result.exit_code == 0

    @patch("hopla.cli.get_group.HabiticaGroupRequest.default_headers")
    @patch("hopla.cli.get_group.requests.get")
    def test_get_group_group_id_ok(self,
                                   mock_get: MagicMock,
                                   headers: MagicMock):
        """A successful hopla get-group call with a group_id."""
        group_id = "eeeeeeee-683b-4b8a-9ddd-b7b652470bdd"
        party_data = {"quest": {"active": False}}
        mock_get.return_value = MockGroupResponse(json={"success": True,
                                                        "data": party_data})
        # No need for credentials, were mocking the API call
        headers.return_value = {"mock": "headers"}

        runner = CliRunner()
        result: Result = runner.invoke(get_group, [group_id])

        mock_get.assert_called_once()
        assert mock_get.call_args.url.endswith(f"/group/{group_id}")
        assert result.exit_code == 0
        assert '"quest": {' in result.stdout
        assert '"active": false' in result.stdout

    @patch("hopla.cli.get_group.HabiticaGroupRequest.default_headers")
    @patch("hopla.cli.get_group.requests.get")
    def test_get_group_404fail(self,
                               mock_get: MagicMock,
                               headers: MagicMock):
        """A failed hopla get-group call."""
        group_id = "fake-uuid"

        err_msg = "NotFound: could not find this group"
        mock_get.return_value = MockGroupResponse(json={
            "success": False, "message": err_msg, "status_code": 404
        })
        headers.return_value = {"mock": "headers"}
        runner = CliRunner()
        result: Result = runner.invoke(get_group, [group_id])

        mock_get.assert_called_once()
        assert mock_get.call_args.url.endswith(f"/group/{group_id}")

        assert result.exit_code == 1
        assert err_msg in result.stdout


def get_test_party_dict() -> dict:
    """Returns pseudo-realistic json as a dict for a party."""
    group_id = "eeeeeeee-683b-4b8a-9ddd-b7b652470bdd"
    group_leader_id = "dddddddd-9e07-434e-8ab4-aaaaaaaaaaaa"
    group_member_id = "eeeeeeee-79a2-4dbe-b8a8-bbbbbbbbbbbb"
    member_username = "melvio"
    leader_username = "maxoya"

    chat = get_test_chat_msgs(
        group_id=group_id,
        group_leader_id=group_leader_id,
        group_member_id=group_member_id,
        leader_username=leader_username,
        member_username=member_username
    )

    return {
        "leaderOnly": {"challenges": True, "getGems": False},
        "quest": {
            "progress": {"collect": {}},
            "active": False,
            "members": {group_leader_id: None, group_member_id: True},
            "extra": {},
            "key": "taskwoodsTerror1",
            "leader": group_leader_id
        },
        "tasksOrder": {"habits": [], "dailys": [], "todos": [], "rewards": []},
        "purchased": {},
        "privacy": "private",
        "chat": chat,
        "memberCount": 2,
        "challengeCount": 0,
        "balance": 0,
        "_id": group_leader_id,
        "summary": "Gut aussehen, Examen rocken",
        "type": "party",
        "name": "best party",
        "description": "Our Party!",
        "managers": {},
        "categories": [],
        "leader": {
            "auth": {"local": {"username": leader_username}},
            "flags": {"verifiedUsername": True},
            "profile": {"name": leader_username},
            "_id": group_id,
            "id": group_id
        },
        "id": group_id
    }


def get_test_chat_msgs(group_id: str, group_leader_id: str, group_member_id: str,
                       leader_username: str, member_username: str) -> list:
    """Returns pseudo-realistic chat json as a dict for a party."""
    chat_msg1 = {
        "flagCount": 0,
        "_id": group_leader_id,
        "flags": {},
        "id": group_leader_id,
        "text": "14000 perception... such overkill .. ",
        "unformattedText": "14000 perception... such overkill .. ",
        "info": {},
        "timestamp": 1632039168020,
        "likes": {},
        "client": "web",
        "uuid": group_leader_id, "contributor": {},
        "backer": {},
        "user": leader_username,
        "username": leader_username,
        "groupId": group_id,
        "userStyles": {
            "items": {
                "gear": {
                    "costume": {
                        "armor": "armor_armoire_vernalVestment",
                        "head": "head_armoire_vernalHennin", "shield": "shield_armoire_redRose",
                        "weapon": "weapon_base_0", "body": "body_base_0",
                        "eyewear": "eyewear_base_0", "back": "back_base_0",
                        "headAccessory": "headAccessory_base_0"},
                    "equipped": {
                        "armor": "armor_warrior_5", "head": "head_special_2",
                        "shield": "shield_special_goldenknight", "weapon": "weapon_warrior_6",
                        "eyewear": "eyewear_special_aetherMask",
                        "body": "body_special_aetherAmulet", "back": "back_special_aetherCloak"}
                },
                "currentMount": "Axolotl-CottonCandyPink", "currentPet": ""
            },
            "preferences": {
                "hair": {"color": "brown", "base": 0, "bangs": 3, "beard": 0, "mustache": 0,
                         "flower": 0},
                "skin": "f5a76e", "shirt": "blue", "chair": "none", "size": "slim",
                "background": "blue", "costume": True
            },
            "stats": {
                "class": "warrior",
                "buffs": {"seafoam": False, "shinySeed": False, "spookySparkles": False,
                          "snowball": False}
            }
        }
    }
    chat_msg2 = {
        "flagCount": 0,
        "_id": group_member_id,
        "flags": {},
        "id": group_member_id,
        "text": f"`{member_username} casts Ethereal Surge for the party.`",
        "unformattedText": f"{member_username} casts Ethereal Surge for the party.",
        "info": {"type": "spell_cast_party", "user": member_username, "class": "wizard",
                 "spell": "mpheal"},
        "timestamp": 1632038593015,
        "likes": {},
        "uuid": "system",
        "groupId": group_id
    }
    return [chat_msg1, chat_msg2]



"""
A modules with application wide error handling logic.
"""
from dataclasses import dataclass

import click
from hopla.hoplalib.common import GlobalConstants


class YouFoundABugRewardError(click.ClickException):
    """An exception that helps users submit a bug report.

    Devnote: This is a defensive programming technique.
    """

    def __init__(self, message):
        super().__init__(message=self._append_issue_link(message))

    @staticmethod
    def _append_issue_link(message):
        return (f"{message}\nYou found a bug! Impressive work.\n"
                "Our hardworking volunteers can go and fix it, after \n"
                "you report it over here:\n "
                f"{GlobalConstants.NEW_ISSUE_URL}")


@dataclass
class PrintableException(BaseException):
    """A BaseException that implements a default __str__ and __repr__."""

    def __init__(self, msg: str):
        super().__init__(PrintableException.__class__)
        self.msg = msg

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.msg})"

    def __str__(self) -> str:
        return self.msg



#!/usr/bin/env python3
"""
Module with code for the `hopla hatch quest-egg` command.
"""
import logging

import click

from hopla.cli.groupcmds.hatch import hatch_egg
from hopla.hoplalib.hatchery.egg_data import EggData
from hopla.hoplalib.hatchery.hatchpotion_data import HatchPotionData

log = logging.getLogger()


@click.command()
@click.argument(
    "egg_name", metavar="QUEST_EGG_NAME",
    type=click.Choice(EggData.quest_egg_names)
)
@click.argument(
    "potion_name",
    type=click.Choice(HatchPotionData.drop_hatch_potion_names)
)
def quest_egg(egg_name: str, potion_name: str):
    """hatch a quest egg.

    \b
    QUEST_EGG_NAME =
        Name of the quest egg. (e.g. Frog, Cow, Spider, TRex, Dolphin)
    \b
    POTION_NAME =
        Name of the potion. Must be a standard drop hatching potion.

    \b
    Examples
    -----
    # hatch a Monkey-Shade
    $ hopla hatch quest-egg Monkey Shade

    \b
    # hatch a Sabretooth-Base
    $ hopla hatch quest-egg Sabretooth Base

    """
    log.debug(f"hopla hatch quest-egg {egg_name=} {potion_name=}")

    hatch_egg(egg_name=egg_name, potion_name=potion_name)



#!/usr/bin/env python3
from datetime import datetime, date, time, timezone, timedelta
from typing import Any, Callable, List
from unittest.mock import MagicMock, patch

import pytest
from requests.structures import CaseInsensitiveDict

from hopla.hoplalib.http import ResponseHeaders
from hopla.hoplalib.throttling import ApiRequestThrottler, RateLimitingAwareThrottler


def is_close(a: float, b: float, epsilon=1e-3) -> bool:
    return abs(a - b) < epsilon


class TestApiRequestThrottler:

    def test__init__(self, ten_api_calls: List[Callable]):
        secs = 3
        call_limit = 12
        throttler = ApiRequestThrottler(ten_api_calls,
                                        throttle_seconds=secs,
                                        throttle_limit=call_limit)

        assert throttler.api_requests == ten_api_calls
        assert throttler.total_calls == len(ten_api_calls)
        assert throttler.throttle_limit == call_limit
        assert throttler.throttle_seconds == 3

    def test_release_without_throttle(self, two_api_requests: List[Callable]):
        throttler = ApiRequestThrottler(two_api_requests)

        call_releaser_result = throttler.release()
        first: Callable = call_releaser_result.__next__()
        second: Callable = call_releaser_result.__next__()
        with pytest.raises(StopIteration):
            call_releaser_result.__next__()

        assert first() == two_api_requests[0]()
        assert second() == two_api_requests[1]()

    def test_execute_all_request_with_throttle(self,
                                               three_api_requests: List[Callable],
                                               capsys):
        secs = 0.001
        throttler = ApiRequestThrottler(three_api_requests,
                                        throttle_limit=1,
                                        throttle_seconds=secs)

        throttler.execute_all_requests()

        captured = capsys.readouterr()
        expected_stderr_msg = \
            f"To prevent overheating Habitica, we'll wait {secs} seconds every request.\n"
        assert captured.err == expected_stderr_msg
        assert captured.out == "some api output"

    @pytest.mark.parametrize(
        "api_request,throttle_limit,expected_should_throttle", [
            ([], 10, False),
            ([lambda: 1, lambda: 2], 1, True),
            ([lambda: 1, lambda: 2], 2, False),
            ([lambda: 1, lambda: 2, lambda: 3], 3, False),
            ([lambda: "a"] * 12, 9, True),
            ([lambda: "a"] * 12, 17, False)
        ]
    )
    def test_exceeds_throttle_limit_noargs(self, api_request: List[Callable],
                                           throttle_limit: int,
                                           expected_should_throttle: bool):
        throttler = ApiRequestThrottler(api_request, throttle_limit=throttle_limit)

        result: bool = throttler.exceeds_throttle_limit()

        assert result == expected_should_throttle

    @pytest.fixture
    def two_api_requests(self) -> List[Callable[[None], Any]]:
        def f1() -> bool:
            return "Thermometer".islower()

        def f2() -> str:
            return "Nebuliser".upper()

        return [f1, f2]

    @pytest.fixture
    def three_api_requests(self, two_api_requests) -> List[Callable]:
        def f3():
            print("some api output", end="")

        return two_api_requests + [f3]

    @pytest.fixture
    def ten_api_calls(self) -> List[Callable[[None], None]]:
        def api_call():
            print("hi")

        return [api_call] * 10


class TestRateLimitingAwareThrottler:
    def test__init__empty_ok(self):
        empty_throttler = RateLimitingAwareThrottler()

        assert empty_throttler.api_requests == []
        assert empty_throttler._is_rate_initialized is False
        assert empty_throttler._xrate_limit_remaining is None
        assert empty_throttler._xrate_limit_reset is None
        assert empty_throttler._api_requests_remaining == 0

    def test_perform_and_yield_response_no_requests_ok(self):
        empty_throttler = RateLimitingAwareThrottler()

        response_generator = empty_throttler.perform_and_yield_response()

        with pytest.raises(StopIteration):
            next(response_generator)

    @patch("hopla.hoplalib.throttling.datetime")
    def test__calculate_sleep_time_wait_till_limit_ok(self, mock_datetime: MagicMock):
        # Testing internals here, feel free to improve.
        # This does circumvent actually 'sleeping' during the
        # tests, but time.sleep can be mocked too!
        throttler = RateLimitingAwareThrottler()
        throttler._xrate_limit_remaining = 0
        reset_datetime = datetime(year=2069, month=10, day=1,
                                  hour=0, minute=0, second=30,
                                  tzinfo=timezone.utc)
        throttler._xrate_limit_reset = reset_datetime

        secs_till_reset = 60
        mock_datetime.now.return_value = reset_datetime - timedelta(seconds=secs_till_reset)

        sleep_result_secs: float = throttler._calculate_sleep_time()

        expected_sleep: float = secs_till_reset + throttler.leeway_seconds + 5
        assert is_close(sleep_result_secs, expected_sleep)

    @pytest.mark.parametrize("secs_till_reset", [60, 40, 15, 10, 6.7])
    @patch("hopla.hoplalib.throttling.datetime")
    def test__calculate_sleep_time_requests_remaining_ok(self, mock_datetime: MagicMock,
                                                         secs_till_reset: float):
        # Testing internals here, feel free to improve.
        # This does circumvent actually 'sleeping' during the
        # tests, but time.sleep can be mocked too!
        throttler = RateLimitingAwareThrottler()
        xrate_limit_remaining = 15
        throttler._xrate_limit_remaining = xrate_limit_remaining
        reset_datetime = datetime(year=2069, month=10, day=1,
                                  hour=0, minute=0, second=30,
                                  tzinfo=timezone.utc)
        throttler._xrate_limit_reset = reset_datetime

        mock_datetime.now.return_value = reset_datetime - timedelta(seconds=secs_till_reset)

        sleep_result_secs: float = throttler._calculate_sleep_time()

        expected_sleep_without_leeway = secs_till_reset / xrate_limit_remaining
        expected_sleep: float = expected_sleep_without_leeway + throttler.leeway_seconds
        assert is_close(sleep_result_secs, expected_sleep)

    @patch("hopla.hoplalib.throttling.Response")
    def test_perform_and_yield_response_single_requests_ok(self,
                                                           mock_response: MagicMock):
        xrate_remaining = "29"
        xrate_reset = "Mon Oct 16 2022 13:49:39 GMT+0000 (Coordinated Universal Time)"
        headers = {
            ResponseHeaders.XRATE_LIMIT_REMAINING_HEADER_NAME: xrate_remaining,
            ResponseHeaders.XRATE_LIMIT_RESET_HEADER_NAME: xrate_reset
        }
        mock_response.headers = CaseInsensitiveDict(data=headers)
        throttler = RateLimitingAwareThrottler([(lambda: mock_response)])

        response_generator = throttler.perform_and_yield_response()

        assert throttler._is_rate_initialized is False
        assert throttler._api_requests_remaining == 1

        assert mock_response is next(response_generator)
        assert throttler._api_requests_remaining == 0

        assert throttler._is_rate_initialized is True
        assert throttler._xrate_limit_remaining == int(xrate_remaining)

        reset_datetime: datetime = throttler._xrate_limit_reset
        reset_date: date = reset_datetime.date()
        reset_time: time = reset_datetime.time()
        reset_timezone = reset_datetime.tzinfo
        assert reset_date == date(year=2022, month=10, day=16)
        assert reset_time == time(hour=13, minute=49, second=39)
        assert reset_timezone == timezone.utc

        with pytest.raises(StopIteration):
            next(response_generator)

    @patch("hopla.hoplalib.throttling.Response")
    @patch("hopla.hoplalib.throttling.Response")
    def test_perform_and_yield_response_multiple_requests_ok(self,
                                                             mock_response1: MagicMock,
                                                             mock_response2: MagicMock):
        xrate_remaining1 = "14"
        quick: datetime = datetime.now(timezone.utc) + timedelta(microseconds=100)
        xrate_reset1 = (
                quick.strftime('%a %b %d %Y %H:%M:%S %Z%z')
                + " (Coordinated Universal Time)"
        )
        headers1 = {
            ResponseHeaders.XRATE_LIMIT_REMAINING_HEADER_NAME: xrate_remaining1,
            ResponseHeaders.XRATE_LIMIT_RESET_HEADER_NAME: xrate_reset1
        }
        mock_response1.headers = headers1

        # hop from 14 to 10 is normal when executing concurrently
        xrate_remaining2 = "10"
        headers2 = {
            ResponseHeaders.XRATE_LIMIT_REMAINING_HEADER_NAME: xrate_remaining2,
            ResponseHeaders.XRATE_LIMIT_RESET_HEADER_NAME: xrate_reset1
        }
        mock_response2.headers = headers2

        throttler = RateLimitingAwareThrottler(api_requests=[
            lambda: mock_response1, lambda: mock_response2
        ])

        generator = throttler.perform_and_yield_response()

        assert throttler._api_requests_remaining == 2
        assert throttler._xrate_limit_remaining is None
        assert throttler._xrate_limit_reset is None

        response1 = next(generator)
        assert response1 == mock_response1
        assert throttler._api_requests_remaining == 1
        assert throttler._throttling_required() is False
        assert throttler._xrate_limit_remaining == int(xrate_remaining1)

        response2 = next(generator)
        assert response2 == mock_response2
        assert throttler._api_requests_remaining == 0
        assert throttler._throttling_required() is False
        assert throttler._xrate_limit_remaining == int(xrate_remaining2)

        with pytest.raises(StopIteration):
            next(generator)



#!/usr/bin/env python3
from unittest.mock import MagicMock
import pytest

import requests

from hopla.hoplalib.requests_helper import get_data_or_exit


class TestRequestHelperModule:

    def test_get_data_or_exit_ok(self):
        test_user_data = {"achievements": {"streak": 195}}
        valid_mock_user_dict = {"success": True, "data": test_user_data}

        response = requests.Response()
        response.json = MagicMock(return_value=valid_mock_user_dict)

        result = get_data_or_exit(api_response=response)

        response.json.assert_called_once()
        assert result == test_user_data

    def test_get_data_or_exit_fail(self):
        errmsg = "Don't worry, happens to the best of us."
        error_response = {
            "success": False,
            "error": "BadRequest",
            "message": errmsg,
        }

        response = requests.Response()
        response.json = MagicMock(return_value=error_response)
        response.status_code = 400

        with pytest.raises(SystemExit) as ex:
            get_data_or_exit(response)

        response.json.assert_called_once()
        expected_exit_msg = f"The habitica API call failed: status_code={response.status_code}"
        assert ex.value.code == expected_exit_msg



"""
Module used for building hopla.

[quote](https://setuptools.readthedocs.io/en/latest/setuptools.html):
As PEP 517 is new, support is not universal, and frontends that do
support it may still have bugs. For compatibility, you may want to
put a setup.py file containing only a setuptools.setup() invocation.
"""
import setuptools

setuptools.setup()



#!/usr/bin/env python3

from hopla.cli.request import display_response


class MockOkResponse:
    """
    This mock response was derived from a HTTP GET request to /api/v3/status.
    """

    def json(self):
        return {
            "success": True, "data": {"status": "up"}, "appVersion": "4.234.0"
        }

    @property
    def status_code(self) -> int:
        return 200

    @property
    def headers(self):
        """Most of the headers retrieved from Habitica (but not all). Also changed the date"""
        access_control_allow_headers = (
            "Authorization,Content-Type,Accept,"
            "Content-Encoding,X-Requested-With,x-api-user,x-api-key,x-client"
        )
        access_control_expose_headers = (
            "X-RateLimit-Limit,X-RateLimit-Remaining,"
            "X-RateLimit-Reset,Retry-After"
        )
        return {
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Methods": "OPTIONS,GET,POST,PUT,HEAD,DELETE",
            "Access-Control-Allow-Headers": access_control_allow_headers,
            "Access-Control-Expose-Headers": access_control_expose_headers,
            "X-RateLimit-Limit": 30,
            "X-RateLimit-Remaining": 29,
            "X-RateLimit-Reset": "Mon Oct 16 2022 13:49:39 GMT+0000 (Coordinated Universal Time)",
            "Content-Type": "application/json; charset=utf-8",
            "Content-Length": 62,
            "Vary": "Accept-Encoding",
            "Date": "Mon, 16 Oct 2022 13:48:39 GMT"
        }


class TestDisplayResponse:
    def test_display_response_normal_case(self, capsys):
        response = MockOkResponse()
        display_response(response,
                         show_response_headers=False,
                         show_status_code=False,
                         show_response=True)

        captured = capsys.readouterr()
        assert "RESPONSE HEADERS" not in captured.out
        assert "X-RateLimit-Limit: 30" not in captured.out
        assert "HTTP Status Code: " not in captured.out
        assert captured.out == (
            '{\n'
            '  "success": true,\n'
            '  "data": {\n'
            '    "status": "up"\n'
            '  },\n'
            '  "appVersion": "4.234.0"\n'
            '}\n'
        )
        assert captured.err == ""

    def test_display_response_show_only_headers(self, capsys):
        response = MockOkResponse()
        display_response(response,
                         show_response_headers=True,
                         show_status_code=False,
                         show_response=False)

        captured = capsys.readouterr()
        assert "RESPONSE HEADERS" in captured.out
        assert "X-RateLimit-Limit: 30" in captured.out
        assert "HTTP Status Code: " not in captured.out
        assert ('{\n'
                '  "success": true,\n'
                '  "data": {\n'
                '    "status": "up"\n'
                '  },\n'
                '  "appVersion": "4.234.0"\n'
                '}\n'
                ) not in captured.out
        assert captured.err == ""

    def test_display_response_show_only_status_code(self, capsys):
        response = MockOkResponse()
        display_response(response,
                         show_response_headers=False,
                         show_status_code=True,
                         show_response=False)

        captured = capsys.readouterr()
        assert "RESPONSE HEADERS" not in captured.out
        assert "X-RateLimit-Limit: 30" not in captured.out
        assert all(header not in captured.out for header in response.json().keys())
        assert f"HTTP Status Code: {response.status_code}" in captured.out

        assert ('{\n'
                '  "success": true,\n'
                '  "data": {\n'
                '    "status": "up"\n'
                '  },\n'
                '  "appVersion": "4.234.0"\n'
                '}\n'
                ) not in captured.out
        assert captured.err == ""

    def test_display_response_show_headers_show_content_show_status_ok(self, capsys):
        response = MockOkResponse()
        display_response(response,
                         show_response_headers=True,
                         show_status_code=True,
                         show_response=True)

        captured = capsys.readouterr()
        assert "RESPONSE HEADERS" in captured.out
        assert "X-RateLimit-Limit: 30" in captured.out
        assert "HTTP Status Code: " in captured.out
        assert ('{\n'
                '  "success": true,\n'
                '  "data": {\n'
                '    "status": "up"\n'
                '  },\n'
                '  "appVersion": "4.234.0"\n'
                '}\n'
                ) in captured.out
        assert captured.err == ""

    def test_display_response_show_nothing(self, capsys):
        response = MockOkResponse()
        display_response(response,
                         show_response_headers=False,
                         show_status_code=False,
                         show_response=False)

        captured = capsys.readouterr()
        assert captured.err == ""
        assert captured.out == ""



#!/usr/bin/env python3
"""
Module with models for eggs.
"""
from dataclasses import dataclass, field
from typing import Any, Dict, Iterator

from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.hatchery.egg_data import EggData
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotion


class EggException(YouFoundABugRewardError):
    """Exception raised when there is an error with an Egg."""


@dataclass
class Egg:
    """An Habitica egg."""

    def __init__(self, name: str, *, quantity: int = 1):
        if name not in EggData.egg_names:
            raise EggException(f"{name} is not a valid egg name.")
        if quantity < 0:
            raise EggException(f"{quantity} is below 0.")

        self.name = name
        self.quantity = quantity

    def __eq__(self, other: Any) -> bool:
        """Return True if both eggs have the same name and same quantity."""
        if isinstance(other, Egg) is False:
            return False
        return self.name == other.name and self.quantity == other.quantity

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name}: {self.quantity})"

    def is_standard_egg(self) -> bool:
        """Return True if this is a drop egg. Else False"""
        return self.name in EggData.drop_egg_names

    def is_quest_egg(self) -> bool:
        """Return True if this is a quest egg. Else False"""
        return self.name in EggData.quest_egg_names

    def can_be_hatched_by(self, potion: HatchPotion) -> bool:
        """Return true if the specified potion can hatch this egg."""
        if potion.quantity == 0 or self.quantity == 0:
            return False
        if self.is_quest_egg():
            return potion.is_standard_hatch_potion()
        return True


@dataclass
class EggCollection:
    """Collection of eggs, backed by a Dict[str, Egg]."""

    eggs: Dict[str, int] = field(init=True, compare=False, default_factory=dict)
    __eggs: Dict[str, Egg] = field(init=False)

    def __post_init__(self):
        """Use the given eggs to create __eggs."""
        self.__eggs: Dict[str, Egg] = {
            name: Egg(name, quantity=quantity) for (name, quantity) in self.eggs.items()
        }

    def __len__(self) -> int:
        return len(self.__eggs)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__eggs)

    def __getitem__(self, name: str) -> Egg:
        return self.__eggs[name]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__eggs=})"

    def values(self) -> Iterator[Egg]:
        """Like dict.values but for an EggCollection."""
        for egg in self.__eggs.values():
            yield egg

    def remove_egg(self, egg: Egg) -> "EggCollection":
        """Remove a single eggs from the EggCollection.

        :param egg: egg to remove
        :return: The updated EggCollection
        """
        if self.__eggs.get(egg.name) is None:
            raise EggException(f"{egg.name} was not in the collection {self.__eggs}")
        if self.__eggs[egg.name].quantity == 0:
            raise EggException(f"We had 0 {egg.name} in the collection. Cannot remove any more.")

        self.__eggs[egg.name].quantity -= 1

        return self

    def get_standard_egg_collection(self) -> "EggCollection":
        """Return all the standard eggs in this collection."""
        return EggCollection({
            name: egg.quantity for (name, egg) in self.__eggs.items() if egg.is_standard_egg()
        })

    def get_quest_egg_collection(self) -> "EggCollection":
        """Return all the quest eggs in this collection."""
        return EggCollection({
            name: egg.quantity for (name, egg) in self.__eggs.items() if egg.is_quest_egg()
        })



#!/usr/bin/env python3
"""
The __main__.py module allows hopla to be ran as python -m hopla.
Moreover, it is used by setup.cfg to construct hopla as a console_script.
"""
from . import setup_hopla_application

if __name__ == "__main__":
    setup_hopla_application()



#!/usr/bin/env python3
from unittest.mock import MagicMock, patch

import pytest

from hopla.hoplalib.zoo.petcontroller import FeedPostRequester
from hopla.hoplalib.zoo.zoofeed_algorithms import FeedPlanItem


class TestFeedPostRequester:
    def test_path_is_ok(self):
        pet = "Wolf-Golden"
        food = "Honey"

        feed_requester = FeedPostRequester(pet_name=pet,
                                           food_name=food)

        assert feed_requester.path == f"/user/feed/{pet}/{food}"

    def test_url_is_okay(self):
        pet = "Wolf-Shimmer"
        food = "Meat"

        feed_requester = FeedPostRequester(pet_name=pet,
                                           food_name=food)

        result_url = feed_requester.feed_pet_food_url

        expected_url = f"https://habitica.com/api/v3/user/feed/{pet}/{food}"
        assert result_url == expected_url

    @pytest.mark.parametrize("times,expected_times", [(2, 2), (None, 1)])
    @patch("hopla.hoplalib.zoo.petcontroller.HabiticaRequest.default_headers")
    @patch("hopla.hoplalib.zoo.petcontroller.requests.post")
    def test_post_feed_request(self, mock_post_request: MagicMock,
                               mock_headers: MagicMock,
                               times: int,
                               expected_times: int):
        pet_name = "Egg-CottonCandyPink"
        food_name = "CottonCandyPink"

        feed_requester = FeedPostRequester(
            pet_name=pet_name, food_name=food_name, food_amount=times
        )

        _ = feed_requester.post_feed_request()

        expected_url = f"https://habitica.com/api/v3/user/feed/{pet_name}/{food_name}"
        mock_post_request.assert_called_with(
            url=expected_url,
            headers=mock_headers,
            params={"amount": expected_times},
            timeout=60
        )

    @pytest.mark.parametrize(
        "feed_plan_item", [
            FeedPlanItem(pet_name="Wolf-Ruby", food_name="Chocolate", times=4),
            FeedPlanItem(pet_name="TigerCub-Floral", food_name="Honey", times=2),
            FeedPlanItem(pet_name="Egg-Zombie", food_name="Meat", times=13),
            FeedPlanItem(pet_name="PandaCub-Shade", food_name="Milk", times=1)
        ]
    )
    def test_build_from_feed_plan_item(self, feed_plan_item: FeedPlanItem):
        requester: FeedPostRequester = FeedPostRequester.build_from(feed_plan_item)

        assert requester.pet_name == feed_plan_item.pet_name
        assert requester.food_name == feed_plan_item.food_name
        assert requester.query_params == {"amount": feed_plan_item.times}



#!/usr/bin/env python3
"""
Module that talks to the Habitica API to manage a Habitica user object.
"""
import requests

from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.http import HabiticaRequest, UrlBuilder
from hopla.hoplalib.user.usermodels import HabiticaUser


class HabiticaUserRequest(HabiticaRequest):
    """Class that requests a user model from the Habitica API"""

    def __init__(self):
        self.url = UrlBuilder(path_extension="/user").url

    def request_user(self) -> requests.Response:
        """Perform the user get request and return the response"""
        return requests.get(
            url=self.url,
            headers=self.default_headers,
            timeout=HabiticaRequest.TIMEOUT
        )

    def request_user_data_or_exit(self) -> HabiticaUser:
        """
        Function that request the user from habitica and returns
        a HabiticaUser if the request was successful. Else exits.
        """
        user_response: requests.Response = self.request_user()
        user_data: dict = get_data_or_exit(user_response)
        return HabiticaUser(user_dict=user_data)



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla cast` command.
"""
import logging

import click
import requests

from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.cast.castcontroller import PostCastRequest
from hopla.hoplalib.cast.spellmodel import Spell, SpellData
from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.throttling import ApiRequestThrottler

log = logging.getLogger()


@click.command()
@click.argument("spell_name", type=click.Choice(SpellData.single_arg_spells))
@click.option("--until-out-of-mana", "-u", is_flag=True, default=False,
              help="Keep casting the specified spell until there is insufficient mana left.")
def cast(spell_name: str, until_out_of_mana: bool) -> None:
    """Cast a spell.

    SPELL_NAME the name of the spell to cast

    Currently, there is only support for spells that are not cast on one
    single task.

    \b
    Examples:
    ---
    # Cast heal all.
    $ hopla cast heallAll

    \b
    # Cast tools of trade until you don't have enough mana left to continue casting.
    $ hopla cast toolsOfTrade --until-out-of-mana

    \b
    # Cast earthquake until you don't have enough mana left to continue casting.
    $ hopla cast -u earth

    \f
    [APIDOCS](https://habitica.com/apidoc/#api-User-UserCast)
    :param spell_name: The spell to cast.
    :param until_out_of_mana: Flag that indicates that the spell should be repeated until
     there is insufficient mana left to keep casting the specified spell.
    :return:
    """
    log.debug(f"hopla cast {spell_name=}")

    spell = Spell(spell_name)
    mana: float = cast_spell_or_exit(spell)

    if until_out_of_mana is True:
        times: int = times_until_out_of_mana(spell, remaining_mana=mana)
        dispatcher = ApiRequestThrottler([lambda: cast_spell_or_exit(spell) for _ in range(times)])
        dispatcher.execute_all_requests()


def times_until_out_of_mana(spell: Spell, *, remaining_mana: float) -> int:
    """
    Return how often you can still cast the spell before you run out of mana.
    """
    return int(remaining_mana / spell.mana_required)


def cast_spell_or_exit(spell: Spell) -> float:
    """
    Cast the given spell by executing an API request.
    :param spell:
    :return:
    """
    request = PostCastRequest(spell=spell)
    response: requests.Response = request.post_spell()
    json_data: dict = get_data_or_exit(response)
    mana = HabiticaUser(json_data["user"]).get_mp()
    click.echo(f"{spell.name} casted successfully: {mana:.1f} mana left.")
    return mana



#!/usr/bin/env python3
from typing import List

import pytest

from hopla.hoplalib.hatchery.eggmodels import Egg, EggCollection
from hopla.hoplalib.hatchery.hatchalgorithms import HatchPlan, HatchPlanItem, HatchPlanMaker
from hopla.hoplalib.hatchery.egg_data import EggData
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotion, HatchPotionCollection
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petmodels import Pet


class TestHatchPlanItem:
    def test__repr__ok(self):
        egg_name, potion_name = "Wolf", "Aquatic"
        egg = Egg(egg_name, quantity=2)
        potion = HatchPotion(potion_name, quantity=3)
        plan_item = HatchPlanItem(egg=egg, potion=potion)

        result: str = repr(plan_item)

        assert result == f"HatchPlanItem({egg_name}: {potion_name})"

    @pytest.mark.parametrize("egg,potion", [
        (Egg("Wolf"), HatchPotion("Silver")),
        (Egg("Fox"), HatchPotion("Moonglow")),
        (Egg("Cactus"), HatchPotion("Watery")),
        (Egg("LionCub"), HatchPotion("Shade")),
        (Egg("PandaCub"), HatchPotion("Ruby")),
        (Egg("Whale"), HatchPotion("Base")),
        (Egg("Horse"), HatchPotion("Golden")),
        (Egg("Fox"), HatchPotion("Red"))
    ])
    def test__post_init_ok(self, egg: Egg, potion: HatchPotion):
        try:
            HatchPlanItem(egg=egg, potion=potion)
        except Exception:
            pytest.fail(f"Expected {egg.name} and {potion.name} to match.")

    def test__eq__should_eq_self_ok(self):
        egg_name, potion_name = "Snake", "Base"
        egg = Egg(egg_name)
        potion = HatchPotion(potion_name)
        item = HatchPlanItem(egg=egg, potion=potion)

        assert item == item

    def test__eq__should_ignore_quantity_differences_true(self):
        egg_name, potion_name = "Snake", "Base"

        item_self = HatchPlanItem(
            egg=Egg(egg_name, quantity=2),
            potion=HatchPotion(potion_name, quantity=3)
        )

        item_other = HatchPlanItem(
            egg=Egg(egg_name, quantity=12),
            potion=HatchPotion(potion_name, quantity=103)
        )

        assert item_self == item_other

    def test__eq__different_egg_name_false(self):
        potion_name = "Base"
        item_self = HatchPlanItem(egg=Egg("Snake"), potion=HatchPotion(potion_name))
        item_other = HatchPlanItem(egg=Egg("Horse"), potion=HatchPotion(potion_name))

        assert item_self != item_other

    def test__eq__different_potion_name_false(self):
        egg_name = "Horse"
        item_self = HatchPlanItem(egg=Egg(egg_name), potion=HatchPotion("Base"))
        item_other = HatchPlanItem(egg=Egg(egg_name), potion=HatchPotion("CottonCandyBlue"))

        assert item_self != item_other

    def test__eq__different_thing_false(self):
        item_self = HatchPlanItem(egg=Egg("Horse"), potion=HatchPotion("Base"))

        assert item_self != Egg("Snake")
        assert item_self != HatchPotion("Base")

    def test_format_item(self):
        egg_name = "Snake"
        potion_name = "Base"
        item = HatchPlanItem(egg=Egg(egg_name), potion=HatchPotion(potion_name))

        result: str = item.format_item()

        expected = f"A {egg_name} egg will be hatched by a {potion_name} potion.\n"
        assert result == expected

    @pytest.mark.parametrize("no_clash_pets", [
        Pet("Wolf-Bronze"),
        Pet("Fox-Base"),
        Pet("Wolf-Base", feed_status=FeedStatus(-1)),
        Pet("Wolf-Base", feed_status=FeedStatus(0))
    ])
    def test_can_hatch_with_pet_true(self, no_clash_pets: Pet):
        hatch_item = HatchPlanItem(egg=Egg("Wolf"), potion=HatchPotion("Base"))

        result = hatch_item.can_hatch_with_pet(no_clash_pets)

        assert result is True

    def test_can_hatch_with_pet_false(self):
        clash_pet = Pet("Wolf-Bronze")
        hatch_item = HatchPlanItem(egg=Egg("Wolf"), potion=HatchPotion("Bronze"))

        result = hatch_item.can_hatch_with_pet(clash_pet)

        assert result is False

    @pytest.mark.parametrize("hatch_item", [
        HatchPlanItem(Egg("Wolf"), HatchPotion("Bronze")),
        HatchPlanItem(Egg("Cactus"), HatchPotion("Amber")),
        HatchPlanItem(Egg("Horse"), HatchPotion("Shade"))
    ])
    def test_can_hatch_with_pets_no_pets_always_true(self, hatch_item: HatchPlanItem):
        result = hatch_item.can_hatch_with_pets([])
        assert result is True

    @pytest.mark.parametrize("hatch_item", [
        HatchPlanItem(Egg("Wolf"), HatchPotion("Bronze")),
        HatchPlanItem(Egg("Cactus"), HatchPotion("Amber")),
        HatchPlanItem(Egg("Horse"), HatchPotion("Shade")),
        HatchPlanItem(Egg("Wolf"), HatchPotion("Base")),
    ])
    def test_can_hatch_with_pets_no_clash_true(self, hatch_item: HatchPlanItem):
        no_clash_pets: List[Pet] = [
            Pet("Wolf-Base", feed_status=FeedStatus(0)),
            Pet("Cactus-Shade"),
            Pet("Seahorse-White"),
            Pet("Horse-Shade", feed_status=FeedStatus(-1))
        ]
        result = hatch_item.can_hatch_with_pets(no_clash_pets)
        assert result is True

    @pytest.mark.parametrize("hatch_item", [
        HatchPlanItem(Egg("Wolf"), HatchPotion("Bronze")),
        HatchPlanItem(Egg("Cactus"), HatchPotion("Amber")),
        HatchPlanItem(Egg("Horse"), HatchPotion("Shade")),
        HatchPlanItem(Egg("Wolf"), HatchPotion("Base")),
    ])
    def test_can_hatch_with_pets_clash_false(self, hatch_item: HatchPlanItem):
        clash_pets: List[Pet] = [
            Pet("Wolf-Bronze"), Pet("Cactus-Amber"), Pet("Horse-Shade"), Pet("Wolf-Base")
        ]
        result = hatch_item.can_hatch_with_pets(clash_pets)
        assert result is False


class TestHatchPlan:
    def test__init__(self):
        hatch_plan = HatchPlan()
        assert hatch_plan.plan == []

    def test__eq__empty(self):
        assert HatchPlan() == HatchPlan()

    def test__eq__false(self):
        plan = HatchPlan().add(egg=Egg("Whale"), potion=HatchPotion("Golden"))
        assert plan != HatchPlan()

    def test__repr__empty(self):
        hatch_plan = HatchPlan()

        result: str = repr(hatch_plan)

        assert result == "HatchPlan(empty)"

    def test__repr__not_empty(self):
        hatch_plan = HatchPlan()
        egg_name = "Horse"
        potion_name = "CottonCandyBlue"
        hatch_plan.add(egg=Egg(egg_name), potion=HatchPotion(potion_name))

        result: str = repr(hatch_plan)

        assert result == f"HatchPlan([HatchPlanItem({egg_name}: {potion_name})])"

    def test_format_plan(self):
        plan = HatchPlan() \
            .add(egg=Egg("Fox"), potion=HatchPotion("Dessert")) \
            .add(egg=Egg("Seahorse"), potion=HatchPotion("Desert"))

        result: str = plan.format_plan()

        expected: str = (
            "A Fox egg will be hatched by a Dessert potion.\n"
            "A Seahorse egg will be hatched by a Desert potion.\n"
        )

        assert result == expected

    def test_is_empty_true(self):
        plan = HatchPlan()
        assert plan.is_empty() is True
        assert len(plan) == 0

    def test_is_empty_false(self):
        plan = HatchPlan().add(egg=Egg("Egg"), potion=HatchPotion("Red"))
        assert plan.is_empty() is False
        assert len(plan) == 1

    def test_add_ok(self):
        hatch_plan = HatchPlan()
        egg = Egg("Seahorse")
        potion = HatchPotion("Skeleton")
        hatch_plan.add(egg=egg, potion=potion)

        assert len(hatch_plan) == 1
        assert HatchPlanItem(egg, potion) in hatch_plan

    def test__iter__(self):
        egg1, egg2 = Egg("Wolf"), Egg("Egg")
        ghost_potion = HatchPotion("Ghost")
        base_potion = HatchPotion("Base")
        hatch_plan = HatchPlan() \
            .add(egg=egg1, potion=ghost_potion) \
            .add(egg=egg2, potion=base_potion)

        iter_result = iter(hatch_plan)
        assert next(iter_result) == HatchPlanItem(egg=egg1, potion=ghost_potion)
        assert next(iter_result) == HatchPlanItem(egg=egg2, potion=base_potion)
        with pytest.raises(StopIteration):
            next(iter_result)


class TestHatchPlanMaker:
    def test__repr__(self):
        eggs = EggCollection({"Wolf": 1, "Whale": 2})
        potions = HatchPotionCollection({"Base": 1, "Ember": 2})
        plan_maker = HatchPlanMaker(egg_collection=eggs, hatch_potion_collection=potions, pets=[])

        result: str = repr(plan_maker)

        assert result == (
            f"HatchPlanMaker(\n"
            f"    {eggs},\n"
            f"    {potions},\n"
            f"    plan=HatchPlan(empty)\n"
            ")"
        )

    def test_make_plan_no_eggs(self):
        empty_eggs = EggCollection()
        potions = HatchPotionCollection({"Zombie": 1})

        plan_maker = HatchPlanMaker(
            egg_collection=empty_eggs,
            hatch_potion_collection=potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected_empty_plan = HatchPlan()
        assert result == expected_empty_plan

    def test_make_plan_no_hatch_potions(self):
        eggs = EggCollection({"Owl": 1})
        empty_potions = HatchPotionCollection()
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=empty_potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected_empty_plan = HatchPlan()
        assert result == expected_empty_plan

    def test_make_plan_1potion_1egg_ok(self):
        eggs = EggCollection({"Owl": 1})
        potions = HatchPotionCollection({"Base": 1})
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected = HatchPlan().add(egg=Egg("Owl"), potion=HatchPotion("Base"))
        assert result == expected

    def test_make_plan_1magic_potion_1quest_egg_ok(self):
        eggs = EggCollection({"Owl": 1})
        magic_potions = HatchPotionCollection({"Thunderstorm": 1})
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=magic_potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected_empty_plan = HatchPlan()
        assert result == expected_empty_plan

    def test_make_plan_1magic_potion_2normal_1quest_egg_ok(self):
        eggs = EggCollection({"Owl": 1, "Fox": 2})
        magic_potions = HatchPotionCollection({"Thunderstorm": 1})
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=magic_potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected = HatchPlan().add(egg=Egg("Fox"),
                                   potion=HatchPotion("Thunderstorm"))
        assert result == expected

    def test_make_plan_more_potions_than_eggs(self):
        eggs = EggCollection({"Horse": 1, "Cactus": 2})
        potions = HatchPotionCollection(
            {"SolarSystem": 2, "Zombie": 1, "Base": 1, "Fluorite": 1, "Shade": 3}
        )
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected = HatchPlan() \
            .add(egg=Egg("Horse"), potion=HatchPotion("Zombie")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("SolarSystem")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("Base")) \
            # ^ Can't hatch same Cactus twice with same potion

        assert result == expected

    def test_make_plan_with_pets(self):
        eggs = EggCollection({"Horse": 1, "Cactus": 3})
        potions = HatchPotionCollection(
            {"SolarSystem": 2, "Zombie": 1, "Base": 1, "Fluorite": 1, "Shade": 3}
        )
        pets: List[Pet] = [
            # This pet blocks hatching another Cactus-SolarSystem
            Pet("Cactus-SolarSystem"),
            # No blocking by Horse-Zombie pet. -1 means pet is not available.
            Pet("Horse-Zombie", feed_status=FeedStatus(-1))
        ]
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=potions,
            pets=pets
        )

        result: HatchPlan = plan_maker.make_plan()

        expected = HatchPlan() \
            .add(egg=Egg("Horse"), potion=HatchPotion("Zombie")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("Base")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("Fluorite")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("Shade"))

        assert result == expected

    def test_make_plan_with_lots_of_pets_and_few_potions_ok(self):
        # case emulating the bug from: https://github.com/melvio/hopla/issues/184
        # Here there are lots of pets, lots of eggs, and only 3 vampire potions.
        # There are also only 3 pets available for hatching with those vampire potions.

        # lots of Eggs
        eggs = EggCollection(eggs={egg_name: 20 for egg_name in EggData.egg_names})
        # 3 vampire potions
        potions = HatchPotionCollection(potions={"Vampire": 3})
        # lots of vampire pets, only 3 hatch-able
        hatchable1, hatchable2, hatchable3 = "FlyingPig", "LionCub", "PandaCub"
        pets: List[Pet] = [
            Pet("BearCub-Vampire", feed_status=FeedStatus(5)),
            Pet("Cactus-Vampire", feed_status=FeedStatus(5)),
            Pet("Dragon-Vampire", feed_status=FeedStatus(5)),
            Pet(hatchable1 + "-Vampire", feed_status=FeedStatus(-1)),
            Pet("Fox-Vampire", feed_status=FeedStatus(5)),
            Pet(hatchable2 + "-Vampire", feed_status=FeedStatus(-1)),
            Pet(hatchable3 + "-Vampire", feed_status=FeedStatus(-1)),
            Pet("TigerCub-Vampire", feed_status=FeedStatus(5)),
            Pet("Wolf-Vampire", feed_status=FeedStatus(5)),
        ]

        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=potions,
            pets=pets
        )

        plan: HatchPlan = plan_maker.make_plan()
        expected_plan = HatchPlan() \
            .add(egg=Egg(hatchable1), potion=HatchPotion("Vampire")) \
            .add(egg=Egg(hatchable2), potion=HatchPotion("Vampire")) \
            .add(egg=Egg(hatchable3), potion=HatchPotion("Vampire"))

        assert plan == expected_plan



#!/usr/bin/env python3
import pytest

from hopla.hoplalib.cast.spellmodel import Spell, SpellData
from hopla.hoplalib.errors import YouFoundABugRewardError
from tests.hoplalib.cast.test_classmodels import HabiticaClassData


class TestSpellData:

    def test_spell_values_in_single_arg_spells(self):
        assert all(spell in SpellData.single_arg_spells
                   for spell in SpellData.warrior_spells_single_arg)
        assert all(spell in SpellData.single_arg_spells
                   for spell in SpellData.mage_spells_single_arg)
        assert all(spell in SpellData.single_arg_spells
                   for spell in SpellData.healer_spells_single_arg)
        assert all(spell in SpellData.single_arg_spells
                   for spell in SpellData.rogue_spells_single_arg)


class TestSpell:
    def test_class_name_ok(self):
        assert Spell("mpheal").class_name == "wizard"
        assert Spell("heal").class_name == "healer"
        assert Spell("stealth").class_name == "rogue"
        assert Spell("intimidate").class_name == "warrior"

    def test_mana_required_ok(self):
        assert Spell("mpheal").mana_required == 30
        assert Spell("toolsOfTrade").mana_required == 25
        assert Spell("valorousPresence").mana_required == 20
        assert Spell("healAll").mana_required == 25

    def test__eq__(self):
        mpheal_spell = Spell("mpheal")
        assert mpheal_spell == Spell("mpheal")
        assert mpheal_spell != Spell("healAll")

    @pytest.mark.parametrize("spell_name", SpellData.single_arg_spells)
    def test__init__ok(self, spell_name: str):
        spell = Spell(name=spell_name)
        assert spell.name == spell_name
        assert 15 <= spell.mana_required <= 45
        assert spell.class_name in HabiticaClassData.class_names
        assert spell.target_id is None  # not supported for now

    def test__init__fail(self):
        invalid_spell_name = "cureAllDisease"
        with pytest.raises(YouFoundABugRewardError) as execinfo:
            Spell(invalid_spell_name)

        msg = str(execinfo.value)
        assert f"'{invalid_spell_name}' does not exist" in msg



from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner, Result

from hopla.cli.buy.enchanted_armoire import get_buy_times_within_budget, \
    times_until_out_of_gp, enchanted_armoire
from hopla.cli.groupcmds.get_user import HabiticaUser
from tests.testutils.user_test_utils import UserTestUtil


class TestTimesUntilPoor:
    def test_times_until_out_of_gp_budget_is0(self):
        assert times_until_out_of_gp(0.) == 0

    def test_times_until_out_of_gp_budget_is_below100(self):
        assert times_until_out_of_gp(92.3) == 0

    def test_times_until_out_of_gp_budget_is_over100(self):
        assert times_until_out_of_gp(212) == 2


class TestBuyTimesWithBudget:

    @pytest.mark.parametrize(
        "gold,expected_buy_times", [(550.1, 5), (5., 0), (69., 0), (9000.1, 90)]
    )
    def test_get_buy_times_within_budget_out_of_gp(self, gold: float,
                                                   expected_buy_times: int):
        user: HabiticaUser = UserTestUtil.user_with_gp(gold=gold)

        times: int = get_buy_times_within_budget(user=user,
                                                 until_out_of_gp_flag=True,
                                                 requested_times=None)

        assert times == expected_buy_times

    @pytest.mark.parametrize(
        "gold,requested_times,expected_times",
        [(550.1, 5, 5),
         (550.1, 4, 4),
         (5., 1, 0),  # cannot buy 1 time with 5gp
         (700.1, None, 1),  # when requested_times is None, use 1
         (705.5, 9, 7),  # cannot buy 9 times with 705.5 gp
         (742., 4, 4),
         (9000.1, 42, 42)]
    )
    def test_get_buy_times_within_budget_with_times(self, gold: float,
                                                    requested_times,
                                                    expected_times: int):
        user: HabiticaUser = UserTestUtil.user_with_gp(gold=gold)

        times: int = get_buy_times_within_budget(user=user,
                                                 until_out_of_gp_flag=False,
                                                 requested_times=requested_times)

        assert times == expected_times


class TestBuyEnchantedArmoireCliCommand:

    @patch("hopla.cli.buy.enchanted_armoire.BuyEnchantedArmoireRequest.post_buy_request")
    @patch("hopla.cli.buy.enchanted_armoire.HabiticaUserRequest.request_user_data_or_exit")
    def test_buy_enchanted_armoire_ok(self,
                                      user_request: MagicMock,
                                      buy_request: MagicMock):
        gold = 500.
        user_request.return_value = UserTestUtil.user_with_gp(gold=gold)

        class MockBuyResponse:
            def __init__(self, json):
                self._json = json

            def json(self) -> dict:
                return self._json

        exp = 35
        armoire_value = {"type": "experience", "value": exp}
        json_response = {
            "success": True,
            "message": "You wrestle with the Armoire and gain Experience. Take that!",
            "data": {
                "armoire": armoire_value
            }
        }
        buy_request.return_value = MockBuyResponse(json=json_response)

        runner = CliRunner()
        result: Result = runner.invoke(enchanted_armoire)

        assert result.exit_code == 0
        assert '"type": "experience",' in result.stdout
        assert f'"value": {exp}' in result.stdout

    def test_buy_enchanted_armoire_times_and_until_out_of_gp_conflict_fail(self):
        runner = CliRunner()
        result: Result = runner.invoke(enchanted_armoire, ["--times", 10, "--until-out-of-gp"])

        expected_msg = "Error: --times and --until-out-of-gp are mutually exclusive.\n"
        assert result.stdout.endswith(expected_msg)
        assert result.exit_code == 2



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla buy enchanted-armoire` command.
"""
import logging
from dataclasses import dataclass
from typing import Final, Optional

import click

from hopla.cli.groupcmds.get_user import HabiticaUser, HabiticaUserRequest
from hopla.hoplalib.buy.buy_controllers import BuyEnchantedArmoireRequest
from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.hoplalib.throttling import ApiRequestThrottler

log = logging.getLogger()


def buy_from_enchanted_armoire_once() -> None:
    """buy once from the enchanted armoire"""
    buy_request = BuyEnchantedArmoireRequest()

    buy_data: dict = buy_request.post_buy_request_get_data_or_exit()

    enchanted_armoire_award = JsonFormatter(buy_data).format_with_double_quotes()
    click.echo(enchanted_armoire_award)


def times_until_out_of_gp(gp: float) -> int:
    """Returns how often you can buy from the enchanted_armoire given some budget.

    \f
    :param gp:
    :return:
    """
    enchanted_armoire_price = 100
    times: float = gp / enchanted_armoire_price
    return int(times)  # int will round down positive floats for us


class EnchantedArmoireException(click.UsageError):
    """
    Exception raised when there is an error with
    a `hopla buy enchanted-armoire` invocation.
    """


_TIMES_OPT_NAME: Final[str] = "--times"
_UNTIL_OUT_GP_OPT_NAME: Final[str] = "--until-out-of-gp"


@dataclass(frozen=True)
class BuyEnchantedArmoireCommandHelper:
    """Helper class for the `hopla buy enchanted-armoire` command."""
    requested_times: Optional[int]
    until_out_gp_flag: bool

    def __post_init__(self):
        if self.requested_times is not None and self.until_out_gp_flag is True:
            err_msg = f"{_TIMES_OPT_NAME} and {_UNTIL_OUT_GP_OPT_NAME} are mutually exclusive."
            raise EnchantedArmoireException(message=err_msg)


@click.command()
@click.option(
    _TIMES_OPT_NAME, "-t", "requested_times",
    type=click.IntRange(min=0),
    metavar="[TIMES]",
    help="Number of times to buy from the enchanted-armoire.\n"
         f"Note that {_TIMES_OPT_NAME} cannot be combined with {_UNTIL_OUT_GP_OPT_NAME}"

)
@click.option(
    _UNTIL_OUT_GP_OPT_NAME, "-u", "until_out_of_gp_flag",
    is_flag=True, default=False, show_default=True,
    help="Buy from enchanted-armoire until gp runs out.\n"
         f"Note that {_UNTIL_OUT_GP_OPT_NAME} cannot be combined with {_TIMES_OPT_NAME}"

)
def enchanted_armoire(requested_times: Optional[int],
                      until_out_of_gp_flag: bool):
    """Buy from the enchanted armoire

    -t TIMES - the number of times to buy from the enchanted armoire. Must be at
    least 0. When TIMES and -u are omitted, 1 is the default.

    If no options are specified, you buy once.

    \b
    Examples
    ----
    # Buy once from the enchanted armoire.
    $ hopla buy enchanted-armoire

    \b
    # buy 3 times
    $ hopla buy enchanted-armoire -t3
    $ hopla buy enchanted-armoire -t 3
    $ hopla buy enchanted-armoire --times 3
    $ hopla buy enchanted-armoire --times=3

    \b
    # buy until you cannot afford the enchanted-armoire anymore
    $ hopla buy enchanted-armoire --until-out-of-gp
    $ hopla buy enchanted-armoire -u
    """
    log.debug(f"hopla buy enchanted-armoire {requested_times=}, {until_out_of_gp_flag=}")
    BuyEnchantedArmoireCommandHelper(
        requested_times=requested_times, until_out_gp_flag=until_out_of_gp_flag
    )

    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    times: int = get_buy_times_within_budget(user=user,
                                             until_out_of_gp_flag=until_out_of_gp_flag,
                                             requested_times=requested_times)

    buy_requests = [buy_from_enchanted_armoire_once for _ in range(times)]
    dispatcher = ApiRequestThrottler(buy_requests)
    dispatcher.execute_all_requests()


def get_buy_times_within_budget(*, user: HabiticaUser,
                                until_out_of_gp_flag: bool,
                                requested_times: Optional[int]) -> int:
    """Return how often we can buy, given the requested amount and our budget."""
    budget: float = user.get_gp()

    max_times = times_until_out_of_gp(budget)
    if until_out_of_gp_flag:
        wanted_times = max_times
    else:
        wanted_times = requested_times or 1

    times = min(max_times, wanted_times)
    click.echo(f"The current budget is {int(budget)}gp.\n"
               f"Buying: {times} times.", err=True)
    return times



#!/usr/bin/env python3
"""
Module with spell models and data.
"""
from dataclasses import dataclass
from typing import Dict, List, Optional
from uuid import UUID

from hopla.hoplalib.errors import YouFoundABugRewardError


class SpellData:
    """A class with data about spells.

    @see: hopla api content | jq .spells
    """
    SpellName = str
    ManaRequired = int
    HabiticaClass = str
    ClassSpellBook = Dict[SpellName, ManaRequired]
    SpellBook = Dict[HabiticaClass, ClassSpellBook]

    warrior_spells_single_arg: ClassSpellBook = {
        "defensiveStance": 25, "valorousPresence": 20, "intimidate": 15
    }
    """@see: hopla api content | jq .spells.warrior"""

    mage_spells_single_arg: ClassSpellBook = {
        "mpheal": 30, "earth": 35
    }
    """@see: hopla api content | jq .spells.wizard"""

    rogue_spells_single_arg: ClassSpellBook = {
        "toolsOfTrade": 25, "stealth": 45
    }
    """@see: hopla api content | jq .spells.rogue"""

    healer_spells_single_arg: ClassSpellBook = {
        "heal": 15, "brightness": 15, "protectAura": 30, "healAll": 25
    }
    """@see: hopla api content | jq .spells.healer"""

    spell_book_single_arg: SpellBook = {
        "warrior": warrior_spells_single_arg,
        "wizard": mage_spells_single_arg,
        "healer": healer_spells_single_arg,
        "rogue": rogue_spells_single_arg
    }

    single_arg_spells: List[SpellName] = (
            list(warrior_spells_single_arg)
            + list(mage_spells_single_arg)
            + list(rogue_spells_single_arg)
            + list(healer_spells_single_arg)
    )
    """Listing of all spell names that don't require a UUID."""


@dataclass
class Spell:
    """A spell."""
    name: str
    target_id: Optional[UUID] = None
    """target_id not supported for now."""

    def __post_init__(self):
        if self.name not in SpellData.single_arg_spells:
            raise YouFoundABugRewardError(f"Spell '{self.name}' does not exist.")

    @property
    def mana_required(self) -> int:
        """Return how much mana is needed to cast this spell."""
        return SpellData.spell_book_single_arg[self.class_name][self.name]

    @property
    def class_name(self) -> str:
        """Get the Habitica class from the spell name.

        see: https://habitica.fandom.com/wiki/Class_System
        """
        if self.name in SpellData.warrior_spells_single_arg:
            return "warrior"
        if self.name in SpellData.mage_spells_single_arg:
            return "wizard"
        if self.name in SpellData.healer_spells_single_arg:
            return "healer"
        if self.name in SpellData.rogue_spells_single_arg:
            return "rogue"

        err_msg = f"Spell '{self.name}' not implemented yet"
        raise YouFoundABugRewardError(err_msg)  # pragma: no cover
